from pandas import DataFrame, RangeIndex, Series

from eda_report.exceptions import InputError


def clean_column_names(data):
    """Makes sure that *columns/features* have *meaningful* names.

    When an array/sequence/iterable/scalar is used to create a ``DataFrame``
    but no column names are specified, ``pandas`` by default names the columns
    by index as [0, 1, 2, ...] (``RangeIndex``).

    This function renames such columns to ['var_1', 'var_2, 'var_3', ...],
    making references and comparisons much more intuitive.

    :param data: The data whose columns are checked
    :type data: ``pandas.DataFrame``
    """
    # Ensure the data has meaningful column names
    if isinstance(data.columns, RangeIndex):
        data.columns = [f'var_{i+1}' for i in data.columns]

    return data


def validate_multivariate_input(data):
    """Ensures that *multivariate input data* is of type ``pandas.DataFrame``.

    If it isn't, this attempts to explicitly cast it as a ``DataFrame``.

    :param data: The data to process.
    :type data: An array-like, sequence, iterable or dict
    :raises InputError: Raised if the data cannot be converted to a
        ``DataFrame``, as required.
    :return: The data as a pandas ``DataFrame``.
    :rtype: ``pandas.DataFrame``
    """
    if isinstance(data, DataFrame):
        return clean_column_names(data)
    else:
        try:
            # Cast the data as a dataframe
            data = DataFrame(data)
        except Exception:
            raise InputError(
                f'Expected a pandas.Dataframe object, but got {type(data)}.'
            )
        data = data.infer_objects()
    return clean_column_names(data)


def validate_univariate_input(data):
    """Ensures that *univariate input data* is of type ``pandas.Series``.

    If it isn't, this attempts to explicitly cast it as a ``Series``.

    :param data: The data to process.
    :type data: An array-like, sequence, iterable or dict
    :raises InputError: Raised if the data cannot be converted to a
        ``Series``, as required.
    :return: The data as a pandas ``Series``.
    :rtype: ``pandas.Series``
    """
    if isinstance(data, Series):
        return data
    else:
        try:
            # Cast the data as a series
            data = Series(data)
        except Exception:
            raise InputError(
                f'Expected a pandas.Series object, but got {type(data)}.'
            )
    return data



from pathlib import Path

import pandas as pd

from eda_report.exceptions import InputError


def df_from_file(filename):
    """Reads a *csv* or *excel* file, and loads its contents as a
    ``pandas.DataFrame``.

    This is basically a wrapper around the
    :func:`pandas.read_csv` and :func:`pandas.read_excel` methods
    respectively.

    :param filename: A file name, or the path to a file.
    :type filename: str
    :raises InputError: If the input file is not a valid csv or excel file.
    :return: A pandas ``DataFrame``.
    :rtype: ``pandas.DataFrame``
    """
    file = Path(filename)

    if file.suffix == '.csv':
        return pd.read_csv(file)
    elif file.suffix == '.xlsx':
        return pd.read_excel(file, engine='openpyxl')
    else:
        raise InputError(f'Invalid input file: {filename}')



import pkgutil
from tkinter import Button, Canvas, Frame, PhotoImage, Tk
from tkinter.colorchooser import askcolor
from tkinter.filedialog import askopenfilename, asksaveasfilename
from tkinter.messagebox import showinfo
from tkinter.simpledialog import askstring

from eda_report import get_word_report
from eda_report.read_file import df_from_file

# Load background image and icon
background_image = pkgutil.get_data(__name__, "images/background.png")
icon = pkgutil.get_data(__name__, "images/icon.png")

# Introductory text
description = """
A simple application to help speed up exploratory data analysis and reporting.

Select a file to analyse, and it will be automatically summarised. The result\
 is a report in .docx format, complete with summary statistics and graphs.

"""


def run_in_gui():
    """Starts the *graphical user interface* to the application.

    This provides the entry point for the ``eda_report`` console script
    (command).
    """
    root = Tk()
    root.title('eda_report')
    root.geometry('600x400')
    root.resizable(0, 0)
    root.wm_iconphoto(True, PhotoImage(data=icon))
    EDAGUI(master=root)
    root.mainloop()


class EDAGUI(Frame):
    """This is the blueprint for the `tkinter`_ - based *graphical user
    interface* to the package/application.

    The graphical window is a *tkinter* ``Frame``, with a *button* that
    opens a *file-dialog* to navigate to and select a file to analyse.

    Once a file is selected, a *text-input widget* and *color-picker tool* pop
    up to help set the report's *title* and *graph color* respectively.
    Finally, a file-dialog to set the desired location and name for the report
    is shown.

    The collected *input file-path*, *title*, *color* and *output file-path*
    are then passed to the :func:`~eda_report.get_word_report` function to
    create the exploratory data analysis report document. A message-box will
    be shown when the report is ready, or when an exception occurs.

    .. _`tkinter`: https://docs.python.org/3/library/tkinter.html
    """

    def __init__(self, master=None, **kw):
        super().__init__(master)
        self.master = master
        self.configure(height=400, width=600)
        self._create_widgets()
        self.pack()

    def _create_widgets(self):
        """Creates the widgets for the graphical user interface: A *canvas*
        with the a *background image*, *introductory text*, and a *button* to
        select a file.
        """
        # Create the canvas
        self.canvas = Canvas(self, width=600, height=400)

        # Set background image
        self.bg_image = PhotoImage(data=background_image)

        self.canvas.create_image(0, 0, image=self.bg_image, anchor="nw")

        # Add introductory text
        self.canvas.create_text(180, 80, text='eda_report', width=500,
                                font=("Courier", 35, 'bold'), fill="black")
        self.canvas.create_text(300, 190, text=description, width=500,
                                font=("Times", 13, 'italic'), fill="black")

        # Add a button
        self.button = Button(
            self, text='Select a file', default='active', bg='teal',
            fg='white', width=30, command=self.create_report, relief='flat',
        )
        self.canvas.create_window(170, 270, anchor='nw', window=self.button)

        self.canvas.pack()

    def create_report(self):
        """Collects input from the graphical user interface, and uses the
        :func:`~eda_report.get_word_report` function to generate a report.
        """
        self._get_data_from_file()
        self._get_report_title()
        self._get_graph_color()
        self._get_save_as_name()

        # Generate the report using the provided arguments
        get_word_report(self.data, title=self.report_title,
                        graph_color=self.graph_color,
                        output_filename=self.save_name)
        # Pop up message to declare that the report is finished
        showinfo(message=f'Done! Report saved as {self.save_name!r}.')

    def _get_data_from_file(self):
        """Creates a file dialog to help navigate to and select a file to
        analyse.
        """
        file_name = askopenfilename(
            title='Select a file to analyse',
            filetypes=(('csv', '*.csv'), ('excel', '*.xlsx'))
        )
        if not file_name:
            showinfo(message='Please select a file to continue')
            file_name = self._get_data_from_file()
        else:
            self.data = df_from_file(file_name)

    def _get_report_title(self):
        """Creates a simple dialog to capture text input for the desired
        report title.
        """
        # Prompt user for report title
        report_title = askstring(
            title='Report Title',
            prompt='Please enter your preferred title for the report:',
            initialvalue='Exploratory Data Analysis Report')

        self.report_title = report_title if report_title else \
            'Exploratory Data Analysis Report'

    def _get_graph_color(self):
        """Creates a graphical color picking tool to help set the desired
        color for the generated graphs.
        """
        # Returns a tuple e.g ((255.99609375, 69.26953125, 0.0), '#ff4500').
        color = askcolor(
            color='orangered',
            title='Please select a color for the graphs'
        )[-1]
        self.graph_color = color if color is not None else 'orangered'

    def _get_save_as_name(self):
        """Create a file dialog to help select a destination folder and file
        name for the generated report.
        """
        # Propmt user for desired output-file name
        save_name = asksaveasfilename(
            initialfile='eda-report.docx', filetypes=(('.docx', '*.docx'),),
            title='Please select Save As file name'
        )
        self.save_name = save_name if save_name else 'eda-report.docx'



class Error(Exception):
    """The base class for exceptions in this package."""
    pass


class InputError(Error):
    """The *Exception* raised when a given input object is *not of the
    expected type* or is otherwise *invalid*.

    In most cases, an attempt is made to cast the erroneous input into the
    proper type, and this *Exception* is raised if it fails.
    """

    def __init__(self, message):
        """Initialise an instance of
        :class:`~eda_report.exceptions.InputError`.

        :param message: A brief description of the mishap detected.
        :type message: str
        """
        self.message = message



from itertools import combinations

import numpy as np
import seaborn as sns
from PIL import Image
from tqdm import tqdm

from eda_report.plotting import Fig, savefig
from eda_report.validate import validate_multivariate_input


class MultiVariable:
    """The blueprint for containers to hold the properties of data with
    *multiple columns/features*.

    The input data is expected to be a ``pandas.DataFrame``. If the data is of
    any other type, then there'll be an attempt to explicitly convert it to a
    ``DataFrame``. If this fails, an :class:`~eda_report.exceptions.InputError`
    is raised.
    """

    def __init__(self, data, *, graph_color='orangered'):
        """Initialise an instance of
        :class:`~eda_report.multivariate.MultiVariable`.

        :param data: The data to process, ideally a ``pandas.DataFrame``
            with several columns.
        :type data: array-like, sequence, iterable, dict
        :param graph_color: The color to apply to the graphs created,
            defaults to 'orangered'. See the *matplotlib* `list of named
            colors`_ for all available options.
        :type graph_color: str, optional

        .. _`list of named colors`:
            https://matplotlib.org/stable/gallery/color/named_colors.html
        """
        self.data = validate_multivariate_input(data)
        #: The color applied to the created graphs.
        self.graph_color = graph_color
        #: A ``DataFrame`` with all the *numeric columns/features* present.
        self.numeric_cols = self._select_cols('number')
        #: A ``DataFrame`` with all the *categorical columns/features*
        #: present. Please note that **boolean features** are also
        #: **considered categorical** in this context.
        self.categorical_cols = self._select_cols('object', 'bool')
        #: A ``DataFrame`` of Pearson correlation coefficients for the
        #: *numeric columns/features*.
        self.correlation_df = self._get_correlation()
        #: Get brief descriptions of the nature of correlation between
        #: numerical feature pairs.
        self.corr_type = {}
        self.bivariate_scatterplots = {}
        self._get_bivariate_analysis()

    def __repr__(self):
        """Create string representation for :class:`Multivariable` objects.
        """
        if self.numeric_cols is None:
            numeric_cols = ''
        else:
            numeric_cols = self.numeric_cols.columns.to_list()

        if self.categorical_cols is None:
            categorical_cols = ''
        else:
            categorical_cols = self.categorical_cols.columns.to_list()

        return f"""\
        Overview
        ========
Numeric features: {', '.join(numeric_cols)}
Categorical features: {', '.join(categorical_cols)}

        Summary Statistics (Numeric features)
        =====================================
{self.numeric_cols.describe() if numeric_cols != '' else 'N/A'}

        Summary Statistics (Categorical features)
        =========================================
{self.categorical_cols.describe() if categorical_cols != '' else 'N/A'}

        Bivariate Analysis (Correlation)
        ================================
{self._corr_description if hasattr(self, '_corr_description') else 'N/A'}
"""

    def show_correlation_heatmap(self):
        """Display a heatmap of Pearson correlation coefficients for all
        *numeric columns/features* present.
        """
        if hasattr(self, 'joint_correlation_plot'):
            image = Image.open(self.joint_correlation_plot)
            image.show()
        else:
            print('Not enough numeric variables to compare.')

    def show_joint_scatterplot(self):
        """Display a joint scatter-plot of all the *numeric columns/features*
        present.
        """
        if hasattr(self, 'joint_scatterplot'):
            image = Image.open(self.joint_scatterplot)
            image.show()
        else:
            print('Not enough numeric variables to compare.')

    def _select_cols(self, *dtypes):
        """Get a DataFrame including only the data types specified.

        :param dtypes: The column data type(s) to include.
        :type dtypes: str
        :return: A ``DataFrame`` with columns of the specified data type(s), or
            ``None`` if no column is of that data type.
        :rtype: A ``DataFrame``, or ``None``
        """
        selected_cols = self.data.select_dtypes(include=dtypes)
        return selected_cols if selected_cols.shape[1] > 0 else None

    def _get_correlation(self):
        """Get a DataFrame of the correlation coefficients for numeric
        columns.
        """
        return None if self.numeric_cols is None else self.data.corr()

    def _get_bivariate_analysis(self):
        """Compare numeric variable pairs.
        """
        if self.numeric_cols is not None and self.numeric_cols.shape[1] > 1:
            self._plot_joint_scatterplot()
            self._plot_joint_correlation()
            self._compare_variable_pairs()
            self._corr_description = '\n'.join(
                [f'{var_pair[0]} & {var_pair[1]} --> {corr_description}'
                 for var_pair, corr_description in self.corr_type.items()])
        else:
            print('Skipped Bivariate Analysis:',
                  'Not enough numeric variables to compare.')

    def _plot_joint_scatterplot(self):
        """Create a joint scatter-plot of all numeric columns.
        """
        fig = sns.pairplot(
            self.numeric_cols,
            plot_kws={'color': self.graph_color},
            diag_kws={'color': self.graph_color}
        )
        fig.fig.suptitle('Scatter-plots of Numeric Columns', size=20)
        self.joint_scatterplot = savefig(fig)

    def _plot_joint_correlation(self):
        """Plot a heatmap of the correlation in all numeric columns."""
        fig = Fig(figsize=(6, 6))
        ax = fig.subplots()
        sns.heatmap(self.correlation_df, annot=True, yticklabels=True,
                    mask=np.triu(self.correlation_df), ax=ax,
                    cmap=sns.light_palette(self.graph_color, as_cmap=True))
        ax.tick_params(rotation=45)
        fig.suptitle('Correlation in Numeric Columns', size=15)

        self.joint_correlation_plot = savefig(fig)

    def _compare_variable_pairs(self):
        """
        Get a brief summary of the nature of correlation between pairs of
        numeric variables.
        """
        self._get_variable_pairs()

        for var1, var2 in tqdm(self.var_pairs, ncols=99,
                               desc='Bivariate analysis'):
            self._quantify_correlation(var1, var2)
            self._regression_plot(var1, var2)

    def _get_variable_pairs(self):
        """Get a list of unique pairings of the numeric variables"""
        self.var_pairs = set(combinations(self.correlation_df.columns, r=2))

    def _quantify_correlation(self, var1, var2):
        """Explain the magnitude of correlation between variable pairs.

        Parameters:
        ----------
        var1, var2: string
            A pair of numeric column(variable) names.
        """
        correlation = self.correlation_df.loc[var1, var2]
        nature = ' positive' if correlation > 0 else ' negative'

        value = abs(correlation)
        if value >= 0.9:
            strength = 'very strong'
        elif value >= 0.7:
            strength = 'strong'
        elif value >= 0.5:
            strength = 'moderate'
        elif value >= 0.3:
            strength = 'weak'
        elif value >= 0.1:
            strength = 'very weak'
        else:
            strength = 'virtually no'
            nature = ''

        self.corr_type[(var1, var2)] = \
            f'{strength}{ nature} correlation ({correlation:.2f})'

    def _regression_plot(self, var1, var2):
        """Create a scatterplot with a fitted linear regression line.

        Parameters:
        ----------
        var1, var2: string
            A pair of numeric column(variable) names.
        """
        fig = Fig(figsize=(8.2, 4))
        ax1, ax2 = fig.subplots(nrows=1, ncols=2)
        sns.regplot(x=var1, y=var2, data=self.data, ax=ax1, truncate=False,
                    color=self.graph_color)
        sns.ecdfplot(data=self.data.loc[:, [var1, var2]], ax=ax2,
                     color=self.graph_color)
        ax1.set_title(f'Scatter-plot - {var1} vs {var2}'.title(), size=9)
        ax2.set_title('Empirical Cummulative Distribution Functions', size=9)

        self.bivariate_scatterplots[(var1, var2)] = savefig(fig)



import logging

from docx import Document
from docx.shared import Inches
from tqdm import tqdm

from eda_report.multivariate import MultiVariable
from eda_report.univariate import Variable


logging.basicConfig(
    format='[%(levelname)s %(asctime)s.%(msecs)03d] %(message)s',
    level=logging.INFO, datefmt='%H:%M:%S'
)


class ReportDocument:
    """The blueprint for objects containing methods to create and populate
    reports in *.docx* format.

    The *input data as a whole* is used to create an instance of
    :class:`~eda_report.multivariate.MultiVariable`, which will compute the
    data's *statistics*, plot *graphs*, and save them all as attributes.

    *Individual columns/features* are likewise processed as instances of
    :class:`~eda_report.univariate.Variable`, so as to obtain *statistics* &
    *plots* for each feature.

    The `python-docx`_ module is then used to write the analysis results as a
    *Word* document. The report is organised into 3 sections:

    #. An *Overview* of the shape of the data and its columns/features
    #. *Univariate Analysis*: Summary statistics and graphs for each column/
        feature.
    #. *Bivariate Analysis*: Pairwise comparisons of all numerical columns.

    .. _python-docx: https://python-docx.readthedocs.io/en/latest/
    """
    def __init__(self, data, *, title='Exploratory Data Analysis Report',
                 graph_color='orangered', output_filename='eda-report.docx',
                 table_style='Table Grid'):
        """Initialise an instance of
        :class:`~eda_report.document.ReportDocument`.

        :param data: The data to analyse.
        :type data: Array-like, sequence, iterable
        :param title: The top level heading for the report, defaults to
            'Exploratory Data Analysis Report'.
        :type title: str, optional
        :param graph_color: The color to apply to the generated graphs,
            defaults to 'orangered'. See the *matplotlib* `list of named
            colors`_ for all available options.
        :type graph_color: str, optional
        :param output_filename: The name and path to use in saving the
            generated report file, defaults to 'eda-report.docx' in the
            current directory.
        :type output_filename: str, optional
        :param table_style: *Microsoft Word* table style to apply to the
            created tables, defaults to 'Table Grid'.
        :type table_style: str, optional

        .. _`list of named colors`:
            https://matplotlib.org/stable/gallery/color/named_colors.html
        """
        self.data = data
        self.TITLE = title
        self.GRAPH_COLOR = graph_color
        self.TABLE_STYLE = table_style
        self.OUTPUT_FILENAME = output_filename
        self.document = Document()
        self._get_report()

    def _get_report(self):
        """Calculate summary statistics, plot graphs, and save the results as
        a .docx file.
        """
        logging.info('Assessing correlation in numeric variables...')
        self.variables = MultiVariable(self.data,
                                       graph_color=self.GRAPH_COLOR)
        logging.info('Done. Summarising each variable...')
        self._create_title_page()  # begin the report document
        self._get_variable_info()  # summarise each variable

        if hasattr(self.variables, 'var_pairs'):
            self._get_bivariate_analysis()  # summarise variable pairs

        self._save_file()
        logging.info(f'Done. Results saved as {self.OUTPUT_FILENAME!r}')

    def _create_title_page(self):
        """Add a title and a brief summary of the data.
        """
        # Title
        self.document.add_heading(self.TITLE, level=0)

        # Brief summary
        num_rows, num_cols = self.data.shape
        num_numeric = self.data.select_dtypes(include='number').columns.size

        if num_rows > 1:
            rows = f'{num_rows} rows (observations)'
        else:
            rows = '1 row (observation)'

        if num_cols > 1:
            cols = f'{num_cols} columns (features)'
        else:
            cols = '1 column (feature)'

        if num_numeric > 1:
            numeric = f', {num_numeric} of which are numeric'
        elif num_numeric == 1:
            numeric = ', 1 of which is numeric'
        else:
            numeric = ''

        intro = f'The dataset consists of {rows} and {cols}{numeric}.'

        self.document.add_paragraph(intro)
        self.document.add_paragraph()

        # Add overview of the data
        self._get_numeric_overview_table()
        self._get_categorical_overview_table()

        if hasattr(self.variables, 'joint_scatterplot'):
            # Add a joint scatterplot of the numeric features
            self.document.add_heading(
                'Joint Scatter-plot of Numeric Columns', level=2)
            self.document.add_picture(self.variables.joint_scatterplot,
                                      width=Inches(6.5))
            self.document.add_page_break()

    def _get_numeric_overview_table(self):
        """Create a table with the overview for numeric features.
        """
        if self.variables.numeric_cols is not None:
            self.document.add_heading('Overview of Numeric Features', level=2)
            self.document.add_paragraph()
            # Get summary statistics for all columns, round the values to 4
            # decimal places, and set the feature names as the index by
            # transposing.
            summary = self.variables.numeric_cols.describe([0.5]).round(4).T
            # Create a table with 9 columns
            self._create_table(
                data=summary, header=True,
                column_widths=(1.5,) + (0.9,)*6, style='Normal Table'
            )

    def _get_categorical_overview_table(self):
        """Create a table with the overview for categorical features.
        """
        if self.variables.categorical_cols is not None:
            self.document.add_heading(
                'Overview of Categorical Features', level=2
            )
            self.document.add_paragraph()
            # Get summary statistics for all columns, round the values to 4
            # decimal places, and set the feature names as the index by
            # transposing.
            summary = self.variables.categorical_cols.describe().round(4).T
            # Create a table with 5 columns
            self._create_table(
                data=summary, header=True,
                column_widths=(1.2,) + (0.8,)*4, style='Normal Table'
            )

    def _get_variable_info(self):
        """Get a brief introduction, summary statistics, and graphs for each
        individual variable.
        """
        self.document.add_heading('A. Univariate Analysis', level=1)
        for idx, col in enumerate(
                tqdm(self.data.columns, ncols=99, desc='Univariate analysis'),
                start=1):
            var = Variable(self.data[col], graph_color=self.GRAPH_COLOR)
            # Heading
            self.document.add_heading(f'{idx}. {col}'.title(), level=2)
            # Introduction
            p = self.document.add_paragraph()
            p.add_run(f'{col}'.capitalize()).bold = True
            p.add_run(f' is a {var.var_type} variable ')
            p.add_run(f'with {var.num_unique} unique values. ')
            p.add_run(f'{var.missing} of its values are missing.')
            # Summary statistics
            self.document.add_heading('Summary Statistics', level=4)
            self._create_table(var.statistics, column_widths=[2.5, 2])
            if hasattr(var, 'most_common_items'):
                self.document.add_heading('Most Common Values', level=4)
                self._create_table(var.most_common_items,
                                   column_widths=(2.5, 2))
            # Graphs
            for graph in var._graphs.values():  # var._graphs is a dict
                self.document.add_picture(graph, width=Inches(5.4))

            self.document.add_page_break()

    def _get_bivariate_analysis(self):
        """Get comparisons and scatterplots for pairs of numeric variables.
        """
        self.document.add_heading(
            'B. Bivariate Analysis (Correlation)', level=1)
        self.document.add_paragraph()
        self.document.add_picture(self.variables.joint_correlation_plot,
                                  width=Inches(6.7))
        self.document.add_page_break()

        for idx, var_pair in enumerate(self.variables.var_pairs, start=1):
            # Heading
            self.document.add_heading(
                f'{idx}. {var_pair[0]} vs {var_pair[1]}'.title(),
                level=2)
            # Introductory text
            p = self.document.add_paragraph()
            p.add_run(f'{var_pair[0]}'.capitalize()).bold = True
            p.add_run(' and ')
            p.add_run(f'{var_pair[1]}'.capitalize()).bold = True
            p.add_run(f' have {self.variables.corr_type[var_pair]}.')
            # Scatter-plot
            self.document.add_picture(
                self.variables.bivariate_scatterplots[var_pair],
                width=Inches(6)
            )
            self.document.add_paragraph()

            if idx % 2 == 0:
                # Add page break after every 2 pairs
                self.document.add_page_break()

    def _create_table(self, data, column_widths=(), style=None, header=False):
        """Create a table from the given data and add it to the document.


        :param data: The data to put into a table.
        :type data: ``pandas.DataFrame``
        :param column_widths: The desired table column widths, defaults to ().
        :type column_widths: tuple, optional
        :param style: *Word* table style to apply, defaults to None.
        :type style: str, optional
        :param header: Whether or not to include column names as the first row
            in the table, defaults to False.
        :type header: bool, optional
        """
        if header:
            # Add the column names as a row.
            data.loc['', :] = data.columns
            # Sort the index in ascending order. Placing the column names at
            # index '' takes them to the top after sorting.
            data.sort_index(inplace=True)

        # Create the table
        table = self.document.add_table(
            rows=len(data),  # A row for each row in the data
            cols=len(column_widths)  # As many columns as column-widths given
        )
        # Set table style
        if style is None:
            table.style = self.document.styles[self.TABLE_STYLE]
        else:
            table.style = self.document.styles[style]

        # Set column dimensions
        for idx, width in enumerate(column_widths):
            table.columns[idx].width = Inches(width)

        # Populate the table with the data
        for idx, row_data in enumerate(data.itertuples()):
            table_row = table.rows[idx]
            # Populate the row's cells
            for cell, datum in zip(table_row.cells, row_data):
                cell.text = f'{datum}'

        self.document.add_paragraph()

    def _save_file(self):
        """Save the document as a .docx file.
        """
        self.document.save(self.OUTPUT_FILENAME)



from io import BytesIO

import matplotlib
from matplotlib.figure import Figure


# Matplotlib configuration
matplotlib.rcParams['figure.dpi'] = 150  # resolution (dots per inch)
matplotlib.rcParams['figure.autolayout'] = True  # tight layout
matplotlib.rcParams['savefig.transparent'] = False  # transparent background
matplotlib.rcParams['savefig.edgecolor'] = 'k'  # black frame
matplotlib.rcParams['savefig.facecolor'] = 'w'  # white background
matplotlib.rcParams['font.family'] = 'serif'
matplotlib.rcParams['axes.spines.top'] = False  # remove top spine
matplotlib.rcParams['axes.spines.right'] = False  # remove left spine
matplotlib.use('agg')  # use non-interactive matplotlib back-end


Fig = Figure


def savefig(figure):
    """Saves the contents of a *matplotlib figure*, in PNG format, as bytes in
    a file-like object.

    This is a utility function helpful in by-passing the *filesystem*. Created
    graphs need not be saved to disk and read from disk; they can instead be
    saved as *attributes* in ``io.BytesIO`` objects, allowing convenient and
    rapid in-memory access.

    :param figure: A *matplotlib figure* with plotted axes.
    :type figure: ``matplotlib.figure.Figure``
    :return: A file-like object with the figure's contents.
    :rtype: ``io.BytesIo``
    """
    graph = BytesIO()
    figure.savefig(graph)
    return graph



from textwrap import shorten

import seaborn as sns
from pandas.api.types import is_bool_dtype, is_numeric_dtype
from PIL import Image
from scipy.stats import probplot

from eda_report.plotting import Fig, savefig
from eda_report.validate import validate_univariate_input


class Variable:
    """This is the blueprint for containers to hold the contents and
    characteristics of *individual columns/features*.
    """

    def __init__(self, data, *, graph_color='orangered', name=None):
        """Initialise an instance of :class:`Variable`.

        :param data: The data to process.
        :type data: ``pandas.Series``
        :param graph_color: The color to apply to the graphs created,
            defaults to 'orangered'.
        :type graph_color: str, optional
        :param name: The feature's name.
        :type name: str, optional
        """
        self.data = validate_univariate_input(data)
        #: The *name* of the *column/feature*. If unspecified in the ``name``
        #: argument during instantiation, this will be taken as the value of
        #: the ``name`` attribute of the input data.
        self.name = self._get_name(name)
        #: The *type* of feature; either *categorical* or *numeric*.
        self.var_type = self._get_variable_type()
        #: *Summary statistics* for the *column/feature*, as a
        #: ``pandas.DataFrame``.
        self.statistics = self._get_summary_statictics()
        #: The *number of unique values* present in the *column/feature*.
        self.num_unique = self.data.nunique()
        #: The set of *unique values* present in the *column/feature*.
        self.unique = set(self.data.unique())
        #: The number of *missing values* (``NaN``, ``None``, ``NA``, ...).
        self.missing = self._get_missing_values()
        #: The *color* applied to the created graphs.
        self.graph_color = graph_color
        # Get graphs for the column/feature as a dict of file-like objects.
        self._graphs = self._plot_graphs()

    def __repr__(self):
        """Creates the string representation for :class:`Variable` objects.
        """
        return f"""\
            Overview
            ========
Name: {self.name},
Type: {self.var_type},
Unique Values: {shorten(f'{self.num_unique} -> {self.unique}', 60)},
Missing Values: {self.missing}

        Summary Statistics
        ==================
{self.statistics}
"""

    def show_graphs(self):
        """Display the plotted graphs for the *column/feature*.
        """
        for graph in self._graphs.values():
            image = Image.open(graph)
            image.show()

    def _get_name(self, name=None):
        """Set the feature's name.

        :param name: The name to give the feature, defaults to None
        :type name: str, optional
        """
        if name:
            self.data = self.data.rename(name)

        return self.data.name

    def _get_variable_type(self):
        """Get the variable type: 'categorical' or 'numeric'.
        """
        if is_numeric_dtype(self.data) and not is_bool_dtype(self.data):
            # Only int and float types
            return 'numeric'
        else:
            # Handle bool, string, datetime, etc as categorical
            self.data = self.data.astype('category')
            return 'categorical'

    def _get_summary_statictics(self):
        """Get summary statistics for the column/feature.
        """
        if self.var_type == 'numeric':
            return self._numeric_summary_statictics()
        elif self.var_type == 'categorical':
            return self._categorical_summary_statictics()

    def _numeric_summary_statictics(self):
        """Get summary statistics for a numeric column/feature.
        """
        summary = self.data.describe()
        summary.index = [
            'Number of observations', 'Average', 'Standard Deviation',
            'Minimum', 'Lower Quartile', 'Median', 'Upper Quartile', 'Maximum'
        ]
        summary['Skewness'] = self.data.skew()
        summary['Kurtosis'] = self.data.kurt()
        return summary.round(7).to_frame()

    def _categorical_summary_statictics(self):
        """Get summary statistics for a categorical column/feature.
        """
        summary = self.data.describe()[['count', 'unique', 'top']]
        summary.index = ['Number of observations', 'Unique values',
                         'Mode (Highest occurring value)']

        # Get most common items and their relative frequency (%)
        most_common_items = self.data.value_counts().head()
        n = len(self.data)
        self.most_common_items = \
            most_common_items.apply(lambda x: f'{x} ({x / n:.2%})').to_frame()

        return summary.to_frame()

    def _get_missing_values(self):
        """Get the number of missing values in the column/feature.
        """
        missing_values = self.data.isna().sum()
        if missing_values == 0:
            return "None"
        else:
            return f"{missing_values} ({missing_values / len(self.data):.2%})"

    def _plot_graphs(self):
        """Plot graphs for the column/feature, based on variable type.
        """
        if self.var_type == 'numeric':
            return {
                'hist_and_boxplot': self._plot_histogram_and_boxplot(),
                'qq_plot': self._plot_qq_plot(),
                'run_plot': self._plot_run_plot()
            }
        elif self.var_type == 'categorical':
            return {
                'bar_plot': self._plot_bar()
            }

    def _plot_histogram_and_boxplot(self):
        """Get a boxplot and a histogram for a numeric column/feature.
        """
        fig = Fig(figsize=(6, 6), linewidth=1)
        ax1, ax2 = fig.subplots(nrows=2, ncols=1)
        # Box-plot
        ax1.boxplot(self.data.dropna(), vert=False, notch=True)
        ax1.set_yticklabels([''])  # Remove y-tick labels
        ax1.set_xlabel(f'{self.name}')
        ax1.set_title(f'Box-plot of {self.name}', size=12)
        # Histogram
        sns.histplot(x=self.data, kde=True, ax=ax2, color=self.graph_color)
        ax2.set_title(f'Distribution plot of {self.name}', size=12)

        return savefig(fig)

    def _plot_qq_plot(self):
        """Get a quantile-quantile probability plot for a numeric column/
        feature.
        """
        # Create a figure and axes
        fig = Fig(figsize=(6, 4), linewidth=1)
        ax = fig.subplots()
        # Get quantile data.
        theoretical_quantiles, ordered_values = probplot(
            self.data,
            fit=False  # The OLS line of best fit will be plotted in regplot
        )
        # Plot the data and a line of best fit
        sns.regplot(x=theoretical_quantiles, y=ordered_values, ax=ax,
                    color=self.graph_color)
        ax.set_title(f'Q-Q Plot (Probability Plot) of {self.name}', size=12)
        ax.set_xlabel('Theoretical Quantiles (~ Standard Normal)')
        ax.set_ylabel('Ordered Values')

        return savefig(fig)

    def _plot_run_plot(self):
        """Get a run-sequence-plot/line-plot for a numeric column/feature.
        """
        # Create a figure and axes
        fig = Fig(figsize=(6, 4), linewidth=1)
        ax = fig.subplots()
        # Get a line plot of the data
        ax.plot(self.data, marker='.', color=self.graph_color)

        xmin = self.data.index[0]
        xmax = self.data.index[-1]
        p50 = self.data.quantile(0.5)
        p5, p95 = self.data.quantile(0.05), self.data.quantile(0.95)
        # Plot a horizontal line at the 50th percentile
        ax.hlines(p50, xmin, xmax, 'grey', '--')
        ax.text(xmax, p50, ' Median')
        # Plot a horizontal line at the 5th percentile
        ax.hlines(p5, xmin, xmax, 'grey', '--')
        ax.text(xmax, p5, ' $5^{th}$ Percentile')
        # Plot a horizontal lines at the 95th percentile
        ax.hlines(p95, xmin, xmax, 'grey', '--')
        ax.text(xmax, p95, ' $95^{th}$ Percentile')

        ax.tick_params(axis='x', rotation=45)  # rotate x-labels by 45°
        ax.set_title(f'Line Plot (Run Plot) of {self.name}', size=12)
        ax.set_ylabel('Observed Value')
        ax.set_xlabel('Index')

        return savefig(fig)

    def _plot_bar(self, color='cyan'):
        """Get a bar-plot for a categorical column/feature.
        """
        fig = Fig(figsize=(6, 4), linewidth=1)
        ax = fig.subplots()

        self.data.value_counts().nlargest(10).plot.bar(
            color=self.graph_color, ax=ax
        )
        ax.tick_params(axis='x', rotation=45)
        ax.set_title(f'Bar-plot of {self.name}', size=12)

        # Annotate bars
        for p in ax.patches:
            ax.annotate(f'{p.get_height():,}', ha='left',
                        xy=(p.get_x(), p.get_height()*1.02),)

        return savefig(fig)



import argparse

from eda_report.read_file import df_from_file


def process_cli_args():
    """Captures and parses input from the command line interface using the
    `argparse`_ module from the Python standard library.

    Available options are::

        usage: eda_cli [-h] [-o OUTFILE] [-t TITLE] [-c COLOR] infile

        Get a basic EDA report in docx format.

        positional arguments:
          infile                A .csv or .xlsx file to process.

        optional arguments:
          -h, --help            show this help message and exit
          -o OUTFILE, --outfile OUTFILE
                                The output file (default: eda-report.docx)
          -t TITLE, --title TITLE
                                The top level heading in the report (default:
                                Exploratory Data Analysis Report)
          -c COLOR, --color COLOR
                                A valid matplotlib color specifier (default:
                                orangered)

    .. _`argparse`: https://docs.python.org/3/library/argparse.html
    """
    parser = argparse.ArgumentParser(
        prog='eda_cli',
        description='Get a basic EDA report in docx format.')

    # Get the data from a file
    parser.add_argument(
        'infile', type=df_from_file,
        help='A .csv or .xlsx file to process.')

    # Set the output file's name
    parser.add_argument(
        '-o', '--outfile', default='eda-report.docx',
        help='The output file (default: %(default)s)')

    # Set the report's title
    parser.add_argument(
        '-t', '--title', default='Exploratory Data Analysis Report',
        help='The top level heading in the report (default: %(default)s)')

    # Set the graph color
    parser.add_argument(
        '-c', '--color', default='orangered',
        help='A valid matplotlib color specifier (default: %(default)s)')

    # Parse the arguments
    return parser.parse_args()



from eda_report.cli import process_cli_args
from eda_report.document import ReportDocument

__version__ = '1.3.0b0'


def get_word_report(data, *, title='Exploratory Data Analysis Report',
                    graph_color='orangered',
                    output_filename='eda-report.docx'):
    """Analyses the input data, and generates a report in *.docx* format.

    This is simply a wrapper around the
    :class:`~eda_report.document.ReportDocument` object, and the arguments
    supplied are passed to it.

    :param data: The data to analyse.
    :type data: an array-like or sequence.
    :param title: The top level heading for the generated report, defaults to
        'Exploratory Data Analysis Report'.
    :type title: str, optional
    :param graph_color: The color to apply to the generated graphs, defaults
        to 'orangered'. See the *matplotlib* `list of named colors`_ for all
        available options.
    :type graph_color: str, optional
    :param output_filename: The name and path to use in saving the generated
        report file, defaults to 'eda-report.docx' in the current directory.
    :type output_filename: str, optional

    .. _`list of named colors`:
        https://matplotlib.org/stable/gallery/color/named_colors.html
    """
    ReportDocument(data, title=title, graph_color=graph_color,
                   output_filename=output_filename)


def run_from_cli():
    """Creates an exploratory data analysis report in *.docx* format using input
    from the command line interface.

    This is the function executed when the package is run as a script (using
    ``python -m eda_report``. It is also the entry point for the ``eda_cli``
    console script (command).

    Arguments passed from the command line are captured using the
    :func:`~eda_report.cli.process_cli_args` function, and then supplied to the
    :func:`~eda_report.get_word_report` function to generate the report.
    """
    args = process_cli_args()
    get_word_report(args.infile, title=args.title, graph_color=args.color,
                    output_filename=args.outfile)
